/***************************************************************************
 *   Copyright (C) 2008 by Volker Lanz <vl@fidra.de>                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA            *
 ***************************************************************************/

#if !defined(FILESYSTEM__H)

#define FILESYSTEM__H

#include <qglobal.h>
#include <QStringList>
#include <QString>
#include <QList>

class Device;
class Report;

class FileSystem
{
	Q_DISABLE_COPY(FileSystem);
	
	public:
		enum Type
		{
			Unknown = 0,
			Extended,

			Ext2,
			Ext3,
			LinuxSwap,
			Fat16,
			Fat32,
			Ntfs,
			ReiserFS,
			Reiser4,
			Xfs,
			Jfs,
			Hfs,
			HfsPlus,
			Ufs,
			Unformatted,
			__lastType
		};

		enum SupportType
		{
			SupportNone = 0,
			SupportInternal = 1,
			SupportLibParted = 2,
			SupportExternal = 4
		};

		Q_DECLARE_FLAGS(SupportTypes, SupportType);

	protected:
		FileSystem(qint64 firstsector, qint64 lastsector, qint64 sectorsused, const QString& label, FileSystem::Type t);

	public:
		virtual ~FileSystem() {}

	public:
		virtual qint64 readUsedCapacity(const QString&) { return -1; }
		virtual QString readLabel(const QString&) { return QString(); }
 		virtual bool create(Report&, const QString&) const { return true; }
 		virtual bool resize(Report&, const QString&, qint64) const { return true; }
		virtual bool writeLabel(Report&, const QString&, const QString&) { return true; }
		virtual bool copy(Report&, const QString&, const QString&) const { return true; }
		virtual bool backup(Report&, const Device&, const QString&, const QString&) const { return false; }
		virtual bool check(Report&, const QString&) const { return true; }
		virtual bool updateUUID(Report&, const QString&) const { return true; }

		virtual SupportType supportGetUsed() const { return SupportNone; }
		virtual SupportType supportGetLabel() const { return SupportNone; }
		virtual SupportType supportCreate() const { return SupportNone; }
		virtual SupportType supportGrow() const { return SupportNone; }
		virtual SupportType supportShrink() const { return SupportNone; }
		virtual SupportType supportMove() const { return SupportNone; }
		virtual SupportType supportCheck() const { return SupportNone; }
		virtual SupportType supportCopy() const { return SupportNone; }
		virtual SupportType supportBackup() const { return SupportNone; }
		virtual SupportType supportSetLabel() const { return SupportNone; }
		virtual SupportType supportUpdateUUID() const { return SupportNone; }

		virtual qint64 minCapacity() const;
		virtual qint64 maxCapacity() const;

		virtual QString name() const;
		virtual FileSystem::Type type() const { return m_Type; }

		static QString nameForType(FileSystem::Type t);
		static QList<FileSystem::Type> types();
		static FileSystem::Type typeForName(const QString& s);
		static FileSystem::Type defaultFileSystem() { return Ext2; }

		virtual QString mountTitle() const;
		virtual QString unmountTitle() const;
		
		qint64 firstSector() const { return m_FirstSector; }
		qint64 lastSector() const { return m_LastSector; }
		qint64 length() const { return lastSector() - firstSector() + 1; }
		
		void setFirstSector(qint64 s) { m_FirstSector = s; }
		void setLastSector(qint64 s) { m_LastSector = s; }

		void move(qint64 newStartSector);
		
		const QString& label() const { return m_Label; }
		qint64 sectorsUsed() const { return m_SectorsUsed; }
		
		void setSectorsUsed(qint64 s) { m_SectorsUsed = s; }
		void setLabel(const QString& s) { m_Label = s; }
		
	protected:
		static bool findExternal(const QString& cmdName, const QStringList& args = QStringList(), int exptectedCode = 1);

	protected:
		FileSystem::Type m_Type;
		qint64 m_FirstSector;
		qint64 m_LastSector;
		qint64 m_SectorsUsed;
		QString m_Label;
};

Q_DECLARE_OPERATORS_FOR_FLAGS(FileSystem::SupportTypes);

#endif
